#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.retry_policy import RetryPolicy


class TestRetryPolicy(unittest.TestCase):

    def test_retry_policy_setting_creation(self):
        retry_policy = RetryPolicy(max_retry_count="1",
                                   strategy="fixed",
                                   delay_interval="5")

        self.assertEqual(retry_policy.get_setting_name(), "retry_policy")
        self.assertEqual(retry_policy.get_dict_repr(),
                         {'setting_name': 'retry_policy',
                          'strategy': 'fixed',
                          'max_retry_count': '1',
                          'delay_interval': '5'})

        retry_policy = RetryPolicy(max_retry_count="1",
                                   strategy="exponential",
                                   minimum_interval="5",
                                   maximum_interval="10")
        self.assertEqual(retry_policy.get_dict_repr(),
                         {'setting_name': 'retry_policy',
                          'strategy': 'exponential',
                          'minimum_interval': '5',
                          'max_retry_count': '1',
                          'maximum_interval': '10'})



import unittest

from azure.functions import DataType
from azure.functions.decorators.core import BindingDirection
from azure.functions.decorators.mcp import MCPToolTrigger
from azure.functions.mcp import MCPToolTriggerConverter
from azure.functions.meta import Datum


class TestMCP(unittest.TestCase):
    def test_mcp_tool_trigger_valid_creation(self):
        trigger = MCPToolTrigger(
            name="context",
            tool_name="hello",
            description="Hello world.",
            tool_properties="[]",
            data_type=DataType.UNDEFINED,
            dummy_field="dummy",
        )
        self.assertEqual(trigger.get_binding_name(), "mcpToolTrigger")
        self.assertEqual(
            trigger.get_dict_repr(),
            {
                "name": "context",
                "toolName": "hello",
                "description": "Hello world.",
                "toolProperties": "[]",
                "type": "mcpToolTrigger",
                "dataType": DataType.UNDEFINED,
                "dummyField": "dummy",
                "direction": BindingDirection.IN,
            },
        )

    def test_trigger_converter(self):
        # Test with string data
        datum = Datum(value='{"arguments":{}}', type='string')
        result = MCPToolTriggerConverter.decode(datum, trigger_metadata={})
        self.assertEqual(result, '{"arguments":{}}')
        self.assertIsInstance(result, str)

        # Test with json data
        datum_json = Datum(value={"arguments": {}}, type='json')
        result_json = MCPToolTriggerConverter.decode(datum_json, trigger_metadata={})
        self.assertEqual(result_json, {"arguments": {}})
        self.assertIsInstance(result_json, dict)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.constants import SERVICE_BUS_TRIGGER, \
    SERVICE_BUS
from azure.functions.decorators.core import BindingDirection, AccessRights, \
    Cardinality, DataType
from azure.functions.decorators.servicebus import ServiceBusQueueTrigger, \
    ServiceBusQueueOutput, ServiceBusTopicTrigger, \
    ServiceBusTopicOutput


class TestServiceBus(unittest.TestCase):
    def test_service_bus_queue_trigger_valid_creation(self):
        trigger = ServiceBusQueueTrigger(name="req", connection="dummy_conn",
                                         queue_name="dummy_queue",
                                         data_type=DataType.UNDEFINED,
                                         access_rights=AccessRights.MANAGE,
                                         is_sessions_enabled=True,
                                         cardinality=Cardinality.ONE,
                                         dummy_field="dummy")

        self.assertEqual(trigger.get_binding_name(), "serviceBusTrigger")
        self.assertEqual(trigger.get_dict_repr(), {
            "type": SERVICE_BUS_TRIGGER,
            "direction": BindingDirection.IN,
            'dummyField': 'dummy',
            "name": "req",
            "connection": "dummy_conn",
            "queueName": "dummy_queue",
            "dataType": DataType.UNDEFINED,
            "accessRights": AccessRights.MANAGE,
            "isSessionsEnabled": True,
            "cardinality": Cardinality.ONE
        })

    def test_service_bus_queue_output_valid_creation(self):
        service_bus_queue_output = ServiceBusQueueOutput(
            name="res",
            connection="dummy_conn",
            queue_name="dummy_queue",
            data_type=DataType.UNDEFINED,
            access_rights=AccessRights.MANAGE,
            dummy_field="dummy")

        self.assertEqual(service_bus_queue_output.get_binding_name(),
                         "serviceBus")
        self.assertEqual(service_bus_queue_output.get_dict_repr(), {
            "type": SERVICE_BUS,
            "direction": BindingDirection.OUT,
            'dummyField': 'dummy',
            "name": "res",
            "dataType": DataType.UNDEFINED,
            "connection": "dummy_conn",
            "queueName": "dummy_queue",
            "accessRights": AccessRights.MANAGE
        })

    def test_service_bus_topic_trigger_valid_creation(self):
        trigger = ServiceBusTopicTrigger(name="req", connection="dummy_conn",
                                         topic_name="dummy_topic",
                                         subscription_name="dummy_sub",
                                         data_type=DataType.UNDEFINED,
                                         access_rights=AccessRights.MANAGE,
                                         is_sessions_enabled=True,
                                         cardinality=Cardinality.ONE,
                                         dummy_field="dummy")

        self.assertEqual(trigger.get_binding_name(), "serviceBusTrigger")
        self.assertEqual(trigger.get_dict_repr(), {
            "type": SERVICE_BUS_TRIGGER,
            "direction": BindingDirection.IN,
            'dummyField': 'dummy',
            "name": "req",
            "connection": "dummy_conn",
            "topicName": "dummy_topic",
            "subscriptionName": "dummy_sub",
            "dataType": DataType.UNDEFINED,
            "accessRights": AccessRights.MANAGE,
            "isSessionsEnabled": True,
            "cardinality": Cardinality.ONE
        })

    def test_service_bus_topic_output_valid_creation(self):
        output = ServiceBusTopicOutput(name="res", connection="dummy_conn",
                                       topic_name="dummy_topic",
                                       subscription_name="dummy_sub",
                                       data_type=DataType.UNDEFINED,
                                       access_rights=AccessRights.MANAGE,
                                       dummy_field="dummy")

        self.assertEqual(output.get_binding_name(), "serviceBus")
        self.assertEqual(output.get_dict_repr(), {
            "type": SERVICE_BUS,
            "direction": BindingDirection.OUT,
            'dummyField': 'dummy',
            "name": "res",
            "dataType": DataType.UNDEFINED,
            "connection": "dummy_conn",
            "topicName": "dummy_topic",
            "subscriptionName": "dummy_sub",
            "accessRights": AccessRights.MANAGE
        })



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import threading
import unittest
from io import StringIO, BytesIO

import pytest

import azure.functions as func
from azure.functions._abc import TraceContext, RetryContext
from azure.functions._http import HttpResponseHeaders
from azure.functions._http_wsgi import (
    WsgiRequest,
    WsgiResponse,
    WsgiMiddleware
)
from azure.functions._http_asgi import AsgiRequest


class WsgiException(Exception):
    def __init__(self, message=''):
        self.message = message


class TestHttpWsgi(unittest.TestCase):

    def test_request_general_environ_conversion(self):
        func_request = self._generate_func_request()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertEqual(environ['REQUEST_METHOD'], 'POST')
        self.assertEqual(environ['SCRIPT_NAME'], '')
        self.assertEqual(environ['PATH_INFO'], '/api/http')
        self.assertEqual(environ['QUERY_STRING'], 'firstname=rt')
        self.assertEqual(environ['CONTENT_TYPE'], 'application/json')
        self.assertEqual(environ['CONTENT_LENGTH'],
                         str(len(b'{ "lastname": "tsang" }')))
        self.assertEqual(environ['SERVER_NAME'], 'function.azurewebsites.net')
        self.assertEqual(environ['SERVER_PORT'], '443')
        self.assertEqual(environ['SERVER_PROTOCOL'], 'HTTP/1.1')

    def test_request_wsgi_environ_conversion(self):
        func_request = self._generate_func_request()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertEqual(environ['wsgi.version'], (1, 0))
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

        self.assertIsInstance(environ['wsgi.input'], BytesIO)
        bytes_io: BytesIO = environ['wsgi.input']
        bytes_io.seek(0)
        self.assertEqual(bytes_io.read(), b'{ "lastname": "tsang" }')

        self.assertIsInstance(environ['wsgi.errors'], StringIO)
        string_io: StringIO = environ['wsgi.errors']
        string_io.seek(0)
        self.assertEqual(string_io.read(), '')

        self.assertEqual(environ['wsgi.multithread'], False)
        self.assertEqual(environ['wsgi.multiprocess'], False)
        self.assertEqual(environ['wsgi.run_once'], False)

    def test_request_http_environ_conversion(self):
        func_request = self._generate_func_request()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertEqual(environ['HTTP_X_MS_SITE_RESTRICTED_TOKEN'], 'xmsrt')

    def test_request_has_no_query_param(self):
        func_request = self._generate_func_request(
            url="https://function.azurewebsites.net",
            params=None)
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertEqual(environ['QUERY_STRING'], '')

    def test_request_has_no_body(self):
        func_request = self._generate_func_request(body=None)
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertEqual(environ['CONTENT_LENGTH'], str(0))

        self.assertIsInstance(environ['wsgi.input'], BytesIO)
        bytes_io: BytesIO = environ['wsgi.input']
        bytes_io.seek(0)
        self.assertEqual(bytes_io.read(), b'')

    def test_request_has_no_headers(self):
        func_request = self._generate_func_request(headers=None)
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertNotIn('CONTENT_TYPE', environ)

    def test_request_protocol_by_header(self):
        func_request = self._generate_func_request(headers={
            "x-forwarded-port": "8081"
        })
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertEqual(environ['SERVER_PORT'], str(8081))
        self.assertEqual(environ['wsgi.url_scheme'], 'https')

    def test_request_protocol_by_scheme(self):
        func_request = self._generate_func_request(url="http://a.b.com")
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)
        self.assertEqual(environ['SERVER_PORT'], str(80))
        self.assertEqual(environ['wsgi.url_scheme'], 'http')

    def test_request_parse_function_context(self):
        func_request = self._generate_func_request()
        func_context = self._generate_func_context()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request,
                              func_context).to_environ(error_buffer)
        self.assertEqual(environ['azure_functions.invocation_id'],
                         '123e4567-e89b-12d3-a456-426655440000'),
        self.assertIsNotNone(environ['azure_functions.thread_local_storage'])
        self.assertEqual(environ['azure_functions.function_name'],
                         'httptrigger')
        self.assertEqual(environ['azure_functions.function_directory'],
                         '/home/roger/wwwroot/httptrigger')
        self.assertEqual(environ['azure_functions.trace_context'],
                         TraceContext)
        self.assertEqual(environ['azure_functions.retry_context'],
                         RetryContext)

    def test_response_from_wsgi_app(self):
        app = self._generate_wsgi_app()
        func_request = self._generate_func_request()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)

        wsgi_response: WsgiResponse = WsgiResponse.from_app(app, environ)
        func_response: func.HttpResponse = wsgi_response.to_func_response()

        self.assertEqual(func_response.mimetype, 'text/plain')
        self.assertEqual(func_response.charset, 'utf-8')
        self.assertEqual(func_response.headers['Content-Type'], 'text/plain')
        self.assertEqual(func_response.status_code, 200)
        self.assertEqual(func_response.get_body(), b'sample string')

    def test_response_no_body(self):
        app = self._generate_wsgi_app(response_body=None)
        func_request = self._generate_func_request()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)

        wsgi_response: WsgiResponse = WsgiResponse.from_app(app, environ)
        func_response: func.HttpResponse = wsgi_response.to_func_response()
        self.assertEqual(func_response.get_body(), b'')

    def test_response_no_headers(self):
        app = self._generate_wsgi_app(response_headers=None)
        func_request = self._generate_func_request()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)

        wsgi_response: WsgiResponse = WsgiResponse.from_app(app, environ)
        func_response: func.HttpResponse = wsgi_response.to_func_response()
        self.assertEqual(func_response.headers, HttpResponseHeaders([]))

    def test_response_with_exception(self):
        app = self._generate_wsgi_app(
            exception=WsgiException(message='wsgi excpt'))
        func_request = self._generate_func_request()
        error_buffer = StringIO()
        environ = WsgiRequest(func_request).to_environ(error_buffer)

        with self.assertRaises(WsgiException) as e:
            wsgi_response = WsgiResponse.from_app(app, environ)
            wsgi_response.to_func_response()

        self.assertEqual(e.exception.message, 'wsgi excpt')

    def test_middleware_handle(self):
        """Test if the middleware can be used by exposing the .handle method,
        specifically when the middleware is used as
        def main(req, context):
            return WsgiMiddleware(app).handle(req, context)
        """
        app = self._generate_wsgi_app()
        func_request = self._generate_func_request()
        func_response = WsgiMiddleware(app).handle(func_request)
        self.assertEqual(func_response.status_code, 200)
        self.assertEqual(func_response.get_body(), b'sample string')

    def test_middleware_handle_with_server_error_status_code(self):
        """Test if the middleware can be used by exposing the .handle method,
        specifically when the middleware is used as
        def main(req, context):
            return WsgiMiddleware(app).handle(req, context)
        """
        app = self._generate_wsgi_app(status="500 Internal Server Error",
                                      response_body=b'internal server error')
        func_request = self._generate_func_request()
        with pytest.raises(Exception) as exec_info:
            func_response = WsgiMiddleware(app).handle(func_request)
            self.assertEqual(func_response.status_code, 500)
        self.assertEqual(exec_info.value.args[0], b'internal server error')

    def test_middleware_wrapper(self):
        """Test if the middleware can be used by exposing the .main property,
        specifically when the middleware is used as
        main = WsgiMiddleware(app).main
        """
        app = self._generate_wsgi_app()
        main = WsgiMiddleware(app).main
        func_request = self._generate_func_request()
        func_context = self._generate_func_context()
        func_response = main(func_request, func_context)
        self.assertEqual(func_response.status_code, 200)
        self.assertEqual(func_response.get_body(), b'sample string')

    def test_path_encoding_utf8(self):
        url = 'http://example.com/Pippi%20L%C3%A5ngstrump'
        request = AsgiRequest(self._generate_func_request(url=url))

        self.assertEqual(request.path_info, u'/Pippi L\u00e5ngstrump')

    def test_path_encoding_latin1(self):
        url = 'http://example.com/Pippi%20L%C3%A5ngstrump'
        request = WsgiRequest(self._generate_func_request(url=url))

        self.assertEqual(request.path_info, u'/Pippi L\u00c3\u00a5ngstrump')

    def _generate_func_request(
            self,
            method="POST",
            url="https://function.azurewebsites.net/api/http?firstname=rt",
            headers={
                "Content-Type": "application/json",
                "x-ms-site-restricted-token": "xmsrt"
            },
            params={
                "firstname": "roger"
            },
            route_params={},
            body=b'{ "lastname": "tsang" }'
    ) -> func.HttpRequest:
        return func.HttpRequest(
            method=method,
            url=url,
            headers=headers,
            params=params,
            route_params=route_params,
            body=body
        )

    def _generate_func_context(
        self,
        invocation_id='123e4567-e89b-12d3-a456-426655440000',
        thread_local_storage=threading.local(),
        function_name='httptrigger',
        function_directory='/home/roger/wwwroot/httptrigger',
        trace_context=TraceContext,
        retry_context=RetryContext
    ) -> func.Context:
        class MockContext(func.Context):
            def __init__(self, ii, tls, fn, fd, tc, rc):
                self._invocation_id = ii
                self._thread_local_storage = tls
                self._function_name = fn
                self._function_directory = fd
                self._trace_context = tc
                self._retry_context = rc

            @property
            def invocation_id(self):
                return self._invocation_id

            @property
            def thread_local_storage(self):
                return self._thread_local_storage

            @property
            def function_name(self):
                return self._function_name

            @property
            def function_directory(self):
                return self._function_directory

            @property
            def trace_context(self):
                return self._trace_context

            @property
            def retry_context(self):
                return self._retry_context

        return MockContext(invocation_id, thread_local_storage,
                           function_name, function_directory,
                           trace_context, retry_context)

    def _generate_wsgi_app(self,
                           status='200 OK',
                           response_headers=[('Content-Type', 'text/plain')],
                           response_body=b'sample string',
                           exception: WsgiException = None):
        class MockWsgiApp:
            _status = status
            _response_headers = response_headers
            _response_body = response_body
            _exception = exception

            def __init__(self, environ, start_response):
                self._environ = environ
                self._start_response = start_response

            def __iter__(self):
                if self._exception is not None:
                    raise self._exception

                self._start_response(self._status, self._response_headers)
                yield self._response_body

        return MockWsgiApp



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Mapping, List
import unittest
import datetime

from azure.functions import meta


class TestMeta(unittest.TestCase):
    def test_parsed_datetime_none(self):
        parsed = self._parse_datetime(None)
        self.assertEqual(parsed, None)

    def test_parse_datetime_empty(self):
        parsed = self._parse_datetime('')
        self.assertEqual(parsed, None)

    def test_utc_datetime_no_fraction_parse(self):
        parsed = self._parse_datetime('2018-12-12T03:16:34Z')
        self.assertEqual(str(parsed), '2018-12-12 03:16:34+00:00')
        self.assertEqual(parsed.tzinfo, datetime.timezone.utc)

        parsed = self._parse_datetime('12/31/9999 23:59:59Z')
        self.assertEqual(str(parsed), '9999-12-31 23:59:59+00:00')
        self.assertEqual(parsed.tzinfo, datetime.timezone.utc)

    def test_utc_datetime_parse(self):
        parsed = self._parse_datetime('2018-12-12T03:16:34.2191Z')
        self.assertEqual(str(parsed), '2018-12-12 03:16:34.219100+00:00')

        parsed = self._parse_datetime('12/31/9999 23:59:59.000000Z')
        self.assertEqual(str(parsed), '9999-12-31 23:59:59+00:00')

        parsed = self._parse_datetime('12/31/9999 23:59:59.999999Z')
        self.assertEqual(str(parsed), '9999-12-31 23:59:59.999999+00:00')

    def test_utc_datetime_neg_tz_parse(self):
        parsed = self._parse_datetime('2018-12-12T03:16:34.2191-00:00')
        self.assertEqual(str(parsed), '2018-12-12 03:16:34.219100+00:00')

        parsed = self._parse_datetime('12/31/9999 23:59:59.999999-00:00')
        self.assertEqual(str(parsed), '9999-12-31 23:59:59.999999+00:00')

    def test_too_fractional_utc_datetime_parse(self):
        parsed1 = self._parse_datetime('2018-12-12T03:16:34.2191989Z')
        self.assertEqual(str(parsed1), '2018-12-12 03:16:34.219198+00:00')

        parsed2 = self._parse_datetime('9999-12-31T23:59:59.9999999+00:00')
        self.assertEqual(str(parsed2), '9999-12-31 23:59:59.999999+00:00')

        parsed3 = self._parse_datetime('12/31/9999 23:59:59.9999999Z')
        self.assertEqual(str(parsed3), '9999-12-31 23:59:59.999999+00:00')

        parsed4 = self._parse_datetime('12/31/9999 23:59:59.9999999+00:00')
        self.assertEqual(str(parsed4), '9999-12-31 23:59:59.999999+00:00')

    def test_local_datetime_no_fraction_parse(self):
        parsed = self._parse_datetime('2018-12-12T03:16:34')
        self.assertEqual(str(parsed), '2018-12-12 03:16:34')

        parsed2 = self._parse_datetime('12/31/9999T23:59:59')
        self.assertEqual(str(parsed2), '9999-12-31 23:59:59')

    def test_local_datetime_parse(self):
        parsed = self._parse_datetime('2018-12-12T03:16:34.2191')
        self.assertEqual(str(parsed), '2018-12-12 03:16:34.219100')

        parsed2 = self._parse_datetime('12/31/9999T23:59:59.219100')
        self.assertEqual(str(parsed2), '9999-12-31 23:59:59.219100')

    def test_too_fractional_local_datetime_parse(self):
        parsed1 = self._parse_datetime('2018-08-07T23:17:57.4610506')
        self.assertEqual(str(parsed1), '2018-08-07 23:17:57.461050')

        parsed2 = self._parse_datetime('9999-12-31T23:59:59.9999999')
        self.assertEqual(str(parsed2), '9999-12-31 23:59:59.999999')

        parsed3 = self._parse_datetime('12/31/9999T23:59:59.9999999')
        self.assertEqual(str(parsed3), '9999-12-31 23:59:59.999999')

    def test_parsed_timedelta_none(self):
        parsed = self._parse_timedelta(None)
        self.assertIsNone(parsed)

    def test_parsed_timedelta_empty(self):
        parsed = self._parse_timedelta('')
        self.assertIsNone(parsed)

    def test_parse_timedelta_seconds(self):
        # Zeros
        parsed = self._parse_timedelta('0')
        self.assertEqual(parsed.seconds, 0)

        # Single Digit
        parsed = self._parse_timedelta('3')
        self.assertEqual(parsed.seconds, 3)

        # Double Digit
        parsed = self._parse_timedelta('56')
        self.assertEqual(parsed.seconds, 56)

        parsed = self._parse_timedelta('678')
        self.assertEqual(parsed.seconds, 678)

    def test_parse_timedelta_minutes_seconds(self):
        # Single Digits Zeros
        parsed = self._parse_timedelta('0:0')
        self.assertEqual(parsed.seconds, 0)

        # Single Digits
        parsed = self._parse_timedelta('3:4')
        self.assertEqual(parsed.seconds, 3 * 60 + 4)

        # Double Digits Zeros
        parsed = self._parse_timedelta('00:00')
        self.assertEqual(parsed.seconds, 0)

        # Double Digits
        parsed = self._parse_timedelta('34:56')
        self.assertEqual(parsed.seconds, 34 * 60 + 56)

    def test_parse_timedelta_hours_minutes_seconds(self):
        # Single Digits Zeros
        parsed = self._parse_timedelta('0:0:0')
        self.assertEqual(parsed.seconds, 0)

        # Single Digits
        parsed = self._parse_timedelta('3:4:5')
        self.assertEqual(parsed.seconds, 3 * 3600 + 4 * 60 + 5)

        # Double Digits Zeros
        parsed = self._parse_timedelta('00:00:00')
        self.assertEqual(parsed.seconds, 0)

        # Double Digits
        parsed = self._parse_timedelta('12:34:56')
        self.assertEqual(parsed.seconds, 12 * 3600 + 34 * 60 + 56)

    def test_parse_utc_datetime_failure(self):
        malformed_utc = '2018-12-12X03:16:34.219289Z'
        with self.assertRaises(ValueError) as context:
            self._parse_datetime(malformed_utc)

        self.assertIn(malformed_utc, str(context.exception))

    def test_parse_local_datetime_failure(self):
        malformed_local = '2018-12-12X03:16:34.219289'
        with self.assertRaises(ValueError) as context:
            self._parse_datetime(malformed_local)

        self.assertIn(malformed_local, str(context.exception))

    def test_datum_single_level_python_value(self):
        datum: Mapping[str, meta.Datum] = meta.Datum(value=None, type="int")
        self.assertEqual(datum.python_value, None)
        self.assertEqual(datum.python_type, type(None))

        datum = meta.Datum(value=1, type=None)
        self.assertEqual(datum.python_value, None)
        self.assertEqual(datum.python_type, type(None))

        datum = meta.Datum(value=b"awesome bytes", type="bytes")
        self.assertEqual(datum.python_value, b"awesome bytes")
        self.assertEqual(datum.python_type, bytes)

        datum = meta.Datum(value="awesome string", type="string")
        self.assertEqual(datum.python_value, 'awesome string')
        self.assertEqual(datum.python_type, str)

        datum = meta.Datum(value=42, type="int")
        self.assertEqual(datum.python_value, 42)
        self.assertEqual(datum.python_type, int)

        datum = meta.Datum(value=43.2103, type="double")
        self.assertEqual(datum.python_value, 43.2103)
        self.assertEqual(datum.python_type, float)

    def test_datum_collections_python_value(self):
        class DatumCollectionString:
            def __init__(self, *args: List[str]):
                self.string = args
        datum = meta.Datum(value=DatumCollectionString("string 1", "string 2"),
                           type="collection_string")
        self.assertListEqual(datum.python_value, ["string 1", "string 2"])
        self.assertEqual(datum.python_type, list)

        class DatumCollectionBytes:
            def __init__(self, *args: List[bytes]):
                self.bytes = args
        datum = meta.Datum(value=DatumCollectionBytes(b"bytes 1", b"bytes 2"),
                           type="collection_bytes")
        self.assertListEqual(datum.python_value, [b"bytes 1", b"bytes 2"])
        self.assertEqual(datum.python_type, list)

        class DatumCollectionSint64:
            def __init__(self, *args: List[int]):
                self.sint64 = args
        datum = meta.Datum(value=DatumCollectionSint64(1234567, 8901234),
                           type="collection_sint64")
        self.assertListEqual(datum.python_value, [1234567, 8901234])
        self.assertEqual(datum.python_type, list)

    def test_datum_json_python_value(self):
        # None
        datum = meta.Datum(value='null',
                           type="json")
        self.assertEqual(datum.python_value, None)
        self.assertEqual(datum.python_type, type(None))

        # Int
        datum = meta.Datum(value='123',
                           type="json")
        self.assertEqual(datum.python_value, 123)
        self.assertEqual(datum.python_type, int)

        # Float
        datum = meta.Datum(value='456.789',
                           type="json")
        self.assertEqual(datum.python_value, 456.789)
        self.assertEqual(datum.python_type, float)

        # String
        datum = meta.Datum(value='"string in json"',
                           type="json")
        self.assertEqual(datum.python_value, "string in json")
        self.assertEqual(datum.python_type, str)

        # List
        datum = meta.Datum(value='["a", "b", "c"]',
                           type="json")
        self.assertListEqual(datum.python_value, ["a", "b", "c"])
        self.assertEqual(datum.python_type, list)

        # Object
        datum = meta.Datum(value='{"name": "awesome", "value": "cool"}',
                           type="json")
        self.assertDictEqual(datum.python_value, {
            "name": "awesome",
            "value": "cool"})
        self.assertEqual(datum.python_type, dict)

        # Should ignore Newlines and Spaces
        datum = meta.Datum(value='{ "name" : "awesome",\n "value":  "cool"\n}',
                           type="json")
        self.assertDictEqual(datum.python_value, {
            "name": "awesome",
            "value": "cool"})
        self.assertEqual(datum.python_type, dict)

    def _parse_datetime(self, datetime_str):
        return meta._BaseConverter._parse_datetime(datetime_str)

    def _parse_timedelta(self, timedelta_str):
        return meta._BaseConverter._parse_timedelta(timedelta_str)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest
from azure.functions.decorators.constants import TABLE
from azure.functions.decorators.table import TableInput, TableOutput
from azure.functions.decorators.core import BindingDirection, DataType


class TestTable(unittest.TestCase):
    def test_table_input_valid_creation(self):
        table_input = TableInput(name="in",
                                 table_name="dummy_table_name",
                                 connection="dummy_in_conn",
                                 row_key="dummy_key",
                                 partition_key="dummy_partition_key",
                                 take=1,
                                 filter="dummy_filter",
                                 data_type=DataType.UNDEFINED)

        self.assertEqual(table_input.get_binding_name(), TABLE)
        self.assertEqual(table_input.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.UNDEFINED,
            "type": "table",
            "name": "in",
            "tableName": "dummy_table_name",
            "connection": "dummy_in_conn",
            "rowKey": "dummy_key",
            "partitionKey": "dummy_partition_key",
            "take": 1,
            "filter": "dummy_filter"
        })

    def test_table_output_valid_creation(self):
        table_output = TableOutput(name="out",
                                   table_name="dummy_table_name",
                                   row_key="dummy_key",
                                   partition_key="dummy_partition_key",
                                   connection="dummy_out_conn",
                                   data_type=DataType.UNDEFINED)

        self.assertEqual(table_output.get_binding_name(), TABLE)
        self.assertEqual(table_output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "dataType": DataType.UNDEFINED,
            "type": "table",
            "name": "out",
            "tableName": "dummy_table_name",
            "connection": "dummy_out_conn",
            "rowKey": "dummy_key",
            "partitionKey": "dummy_partition_key"
        })



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing

from azure.functions import _abc as azf_abc
from . import meta
from ._jsonutils import json


class TimerRequest(azf_abc.TimerRequest):

    def __init__(self, *, past_due: bool = False,
                 schedule_status: typing.Optional[dict] = None,
                 schedule: typing.Optional[dict] = None) -> None:
        self.__past_due = past_due
        self.__schedule_status = schedule_status if schedule_status else {}
        self.__schedule = schedule if schedule else {}

    @property
    def past_due(self) -> bool:
        return self.__past_due

    @property
    def schedule_status(self) -> dict:
        return self.__schedule_status

    @property
    def schedule(self) -> dict:
        return self.__schedule


class TimerRequestConverter(meta.InConverter,
                            binding='timerTrigger', trigger=True):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, azf_abc.TimerRequest)

    @classmethod
    def decode(cls, data: meta.Datum, *, trigger_metadata) -> typing.Any:
        if data.type != 'json':
            raise NotImplementedError

        info = json.loads(data.value)

        return TimerRequest(
            past_due=info.get('IsPastDue', False),
            schedule_status=info.get('ScheduleStatus', {}),
            schedule=info.get('Schedule', {}))



from .extension_meta import ExtensionMeta
from .function_extension_exception import FunctionExtensionException
from .app_extension_base import AppExtensionBase
from .func_extension_base import FuncExtensionBase

__all__ = [
    'ExtensionMeta',
    'FunctionExtensionException',
    'AppExtensionBase',
    'FuncExtensionBase'
]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing
import json

from azure.functions import _durable_functions
from . import meta


# Durable Function Orchestration Trigger
class OrchestrationTriggerConverter(meta.InConverter,
                                    meta.OutConverter,
                                    binding='orchestrationTrigger',
                                    trigger=True):
    @classmethod
    def check_input_type_annotation(cls, pytype):
        return issubclass(pytype, _durable_functions.OrchestrationContext)

    @classmethod
    def check_output_type_annotation(cls, pytype):
        # Implicit output should accept any return type
        return True

    @classmethod
    def decode(cls,
               data: meta.Datum, *,
               trigger_metadata) -> _durable_functions.OrchestrationContext:
        return _durable_functions.OrchestrationContext(data.value)

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        # Durable function context should be a json
        return meta.Datum(type='json', value=obj)

    @classmethod
    def has_implicit_output(cls) -> bool:
        return True


class EnitityTriggerConverter(meta.InConverter,
                              meta.OutConverter,
                              binding='entityTrigger',
                              trigger=True):
    @classmethod
    def check_input_type_annotation(cls, pytype):
        return issubclass(pytype, _durable_functions.EntityContext)

    @classmethod
    def check_output_type_annotation(cls, pytype):
        # Implicit output should accept any return type
        return True

    @classmethod
    def decode(cls,
               data: meta.Datum, *,
               trigger_metadata) -> _durable_functions.EntityContext:
        return _durable_functions.EntityContext(data.value)

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        # Durable function context should be a json
        return meta.Datum(type='json', value=obj)

    @classmethod
    def has_implicit_output(cls) -> bool:
        return True


# Durable Function Activity Trigger
class ActivityTriggerConverter(meta.InConverter,
                               meta.OutConverter,
                               binding='activityTrigger',
                               trigger=True):
    @classmethod
    def check_input_type_annotation(cls, pytype):
        # Activity Trigger's arguments should accept any types
        return True

    @classmethod
    def check_output_type_annotation(cls, pytype):
        # The activity trigger should accept any JSON serializable types
        return True

    @classmethod
    def decode(cls,
               data: meta.Datum, *,
               trigger_metadata) -> typing.Any:
        data_type = data.type

        # Durable functions extension always returns a string of json
        # See durable functions library's call_activity_task docs
        if data_type in ['string', 'json']:
            try:
                callback = _durable_functions._deserialize_custom_object
                result = json.loads(data.value, object_hook=callback)
            except json.JSONDecodeError:
                # String failover if the content is not json serializable
                result = data.value
            except Exception:
                raise ValueError(
                    'activity trigger input must be a string or a '
                    f'valid json serializable ({data.value})')
        else:
            raise NotImplementedError(
                f'unsupported activity trigger payload type: {data_type}')

        return result

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        try:
            callback = _durable_functions._serialize_custom_object
            result = json.dumps(obj, default=callback)
        except TypeError:
            raise ValueError(
                f'activity trigger output must be json serializable ({obj})')

        return meta.Datum(type='json', value=result)

    @classmethod
    def has_implicit_output(cls) -> bool:
        return True


# Durable Functions Durable Client Bindings
class DurableClientConverter(meta.InConverter,
                             meta.OutConverter,
                             binding='durableClient'):
    @classmethod
    def has_implicit_output(cls) -> bool:
        return False



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import collections.abc
import typing

from azure.functions import _sql as sql

from . import meta
from ._jsonutils import json


class SqlConverter(meta.InConverter, meta.OutConverter,
                   binding='sql'):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, sql.BaseSqlRowList)

    @classmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, (sql.BaseSqlRowList, sql.BaseSqlRow))

    @classmethod
    def decode(cls,
               data: meta.Datum,
               *,
               trigger_metadata) -> typing.Optional[sql.SqlRowList]:
        if data is None or data.type is None:
            return None

        data_type = data.type

        if data_type in ['string', 'json']:
            body = data.value

        elif data_type == 'bytes':
            body = data.value.decode('utf-8')

        else:
            raise NotImplementedError(
                f'Unsupported payload type: {data_type}')

        rows = json.loads(body)
        if not isinstance(rows, list):
            rows = [rows]

        return sql.SqlRowList(
            (None if row is None else sql.SqlRow.from_dict(row))
            for row in rows)

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        if isinstance(obj, sql.SqlRow):
            data = sql.SqlRowList([obj])

        elif isinstance(obj, sql.SqlRowList):
            data = obj

        elif isinstance(obj, collections.abc.Iterable):
            data = sql.SqlRowList()

            for row in obj:
                if not isinstance(row, sql.SqlRow):
                    raise NotImplementedError(
                        f'Unsupported list type: {type(obj)}, \
                            lists must contain SqlRow objects')
                else:
                    data.append(row)

        else:
            raise NotImplementedError(f'Unsupported type: {type(obj)}')

        return meta.Datum(
            type='json',
            value=json.dumps([dict(d) for d in data])
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Callable, NamedTuple


class ExtensionHookMeta(NamedTuple):
    """The metadata of a single life-cycle hook.
    The ext_name has the class name of an extension class.
    The ext_impl has the callable function that is used by the worker.
    """
    ext_name: str
    ext_impl: Callable

    # When adding more fields, make sure they have default values (e.g.
    # ext_new_field: Optional[str] = None



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest
import azure.functions.warmup as warmup

from azure.functions.meta import Datum


class TestWarmup(unittest.TestCase):
    def test_warmup_decode(self):
        # given
        datum: Datum = Datum(value='''''', type='json')

        # when
        warmup_context: warmup.WarmUpContext = \
            warmup.WarmUpTriggerConverter.decode(datum, trigger_metadata={})

        # then
        self.assertTrue(isinstance(warmup_context, warmup.WarmUpContext))

    def test_warmup_input_type(self):
        check_input_type = (
            warmup.WarmUpTriggerConverter.check_input_type_annotation
        )
        self.assertTrue(check_input_type(warmup.WarmUpContext))
        self.assertFalse(check_input_type(str))



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.core import BindingDirection
from azure.functions.decorators.constants import DAPR_BINDING, \
    DAPR_BINDING_TRIGGER, DAPR_INVOKE, DAPR_PUBLISH, DAPR_SECRET, \
    DAPR_SERVICE_INVOCATION_TRIGGER, DAPR_STATE, DAPR_TOPIC_TRIGGER
from azure.functions.decorators.function_app import FunctionApp
from tests.utils.testutils import assert_json


class TestDapr(unittest.TestCase):
    def setUp(self):
        self.func_app = FunctionApp()

    def _get_user_function(self, app):
        funcs = app.get_functions()
        self.assertEqual(len(funcs), 1)
        return funcs[0]

    def test_dapr_service_invocation_trigger_default_args(self):
        app = self.func_app

        @app.dapr_service_invocation_trigger(arg_name="req",
                                             method_name="dummy_method_name")
        def test_dapr_service_invocation_trigger_default_args():
            pass

        func = self._get_user_function(app)

        assert_json
        (
            self, func,
            {
                "scriptFile": "function_app.py",
                "bindings": [
                    {
                        "direction": BindingDirection.IN,
                        "type": DAPR_SERVICE_INVOCATION_TRIGGER,
                        "name": "req",
                        "methodName": "dummy_method_name"
                    }
                ]
            }
        )

    def test_dapr_binding_trigger_default_args(self):
        app = self.func_app

        @app.dapr_binding_trigger(arg_name="req",
                                  binding_name="dummy_binding_name")
        def test_dapr_binding_trigger_default_args():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": DAPR_BINDING_TRIGGER,
                                         "name": "req",
                                         "bindingName": "dummy_binding_name"
                                     }
                                 ]
                                 })

    def test_dapr_topic_trigger_default_args(self):
        app = self.func_app

        @app.dapr_topic_trigger(arg_name="req",
                                pub_sub_name="dummy_pub_sub_name",
                                topic="dummy_topic",
                                route="/dummy_route")
        def test_dapr_topic_trigger_default_args():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": DAPR_TOPIC_TRIGGER,
                                         "name": "req",
                                         "pubSubName": "dummy_pub_sub_name",
                                         "topic": "dummy_topic",
                                         "route": "/dummy_route"
                                     }
                                 ]
                                 })

    def test_dapr_state_input_binding(self):
        app = self.func_app

        @app.dapr_service_invocation_trigger(arg_name="req",
                                             method_name="dummy")
        @app.dapr_state_input(arg_name="in",
                              state_store="dummy_state_store",
                              key="dummy_key")
        def test_dapr_state_input_binding():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": DAPR_STATE,
            "name": "in",
            "stateStore": "dummy_state_store",
            "key": "dummy_key"
        })

    def test_dapr_secret_input_binding(self):
        app = self.func_app

        @app.dapr_service_invocation_trigger(arg_name="req",
                                             method_name="dummy")
        @app.dapr_secret_input(arg_name="in",
                               secret_store_name="dummy_secret_store_name",
                               key="dummy_key",
                               metadata="dummy_metadata")
        def test_dapr_secret_input_binding():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": DAPR_SECRET,
            "name": "in",
            "secretStoreName": "dummy_secret_store_name",
            "key": "dummy_key",
            "metadata": "dummy_metadata"
        })

    def test_dapr_state_output_binding(self):
        app = self.func_app

        @app.dapr_service_invocation_trigger(arg_name="req",
                                             method_name="dummy")
        @app.dapr_state_output(arg_name="out",
                               state_store="dummy_state_store",
                               key="dummy_key")
        def test_dapr_state_output_binding():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": DAPR_STATE,
            "name": "out",
            "stateStore": "dummy_state_store",
            "key": "dummy_key"
        })

    def test_dapr_invoke_output_binding(self):
        app = self.func_app

        @app.dapr_service_invocation_trigger(arg_name="req",
                                             method_name="dummy")
        @app.dapr_invoke_output(arg_name="out",
                                app_id="dummy_app_id",
                                method_name="dummy_method_name",
                                http_verb="dummy_http_verb")
        def test_dapr_invoke_output_binding():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": DAPR_INVOKE,
            "name": "out",
            "appId": "dummy_app_id",
            "methodName": "dummy_method_name",
            "httpVerb": "dummy_http_verb"
        })

    def test_dapr_publish_output_binding(self):
        app = self.func_app

        @app.dapr_service_invocation_trigger(arg_name="req",
                                             method_name="dummy")
        @app.dapr_publish_output(arg_name="out",
                                 pub_sub_name="dummy_pub_sub_name",
                                 topic="dummy_topic")
        def test_dapr_publish_output_binding():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": DAPR_PUBLISH,
            "name": "out",
            "pubSubName": "dummy_pub_sub_name",
            "topic": "dummy_topic"
        })

    def test_dapr_binding_output_binding(self):
        app = self.func_app

        @app.dapr_service_invocation_trigger(arg_name="req",
                                             method_name="dummy")
        @app.dapr_binding_output(arg_name="out",
                                 binding_name="dummy_binding_name",
                                 operation="dummy_operation")
        def test_dapr_binding_output_binding():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": DAPR_BINDING,
            "name": "out",
            "bindingName": "dummy_binding_name",
            "operation": "dummy_operation"
        })



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json
import unittest

import azure.functions as func
import azure.functions.mysql as mysql
from azure.functions.meta import Datum


class TestMySql(unittest.TestCase):
    def test_mysql_decode_none(self):
        result: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=None, trigger_metadata=None)
        self.assertIsNone(result)

    def test_mysql_decode_string(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "string")
        result: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'MySqlRowList should be non-None')
        self.assertEqual(len(result),
                         1,
                         'MySqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['id'],
                         '1',
                         'MySqlRow item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test',
                         'MySqlRow item should have name test')

    def test_mysql_decode_bytes(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """.encode(), "bytes")
        result: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'MySqlRowList should be non-None')
        self.assertEqual(len(result),
                         1,
                         'MySqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['id'],
                         '1',
                         'MySqlRow item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test',
                         'MySqlRow item should have name test')

    def test_mysql_decode_json(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "json")
        result: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'MySqlRowList should be non-None')
        self.assertEqual(len(result),
                         1,
                         'MySqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['id'],
                         '1',
                         'MySqlRow item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test',
                         'MySqlRow item should have name test')

    def test_mysql_decode_json_name_is_null(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": null
        }
        """, "json")
        result: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'MySqlRowList itself should be non-None')
        self.assertEqual(len(result),
                         1,
                         'MySqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['name'],
                         None,
                         'Item in MySqlRowList should be None')

    def test_mysql_decode_json_multiple_entries(self):
        datum: Datum = Datum("""
        [
            {
                "id": "1",
                "name": "test1"
            },
            {
                "id": "2",
                "name": "test2"
            }
        ]
        """, "json")
        result: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result),
                         2,
                         'MySqlRowList should have exactly 2 items')
        self.assertEqual(result[0]['id'],
                         '1',
                         'First MySqlRowList item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test1',
                         'First MySqlRowList item should have name test1')
        self.assertEqual(result[1]['id'],
                         '2',
                         'First MySqlRowList item should have id 2')
        self.assertEqual(result[1]['name'],
                         'test2',
                         'Second MySqlRowList item should have name test2')

    def test_mysql_decode_json_multiple_nulls(self):
        datum: Datum = Datum("[null]", "json")
        result: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result),
                         1,
                         'MySqlRowList should have exactly 1 item')
        self.assertEqual(result[0],
                         None,
                         'MySqlRow item should be None')

    def test_mysql_encode_mysqlrow(self):
        mysqlRow = func.MySqlRow.from_json("""
        {
            "id": "1",
            "name": "test"
        }
        """)
        datum = mysql.MySqlConverter.encode(obj=mysqlRow, expected_type=None)
        self.assertEqual(datum.type,
                         'json',
                         'Datum type should be JSON')
        self.assertEqual(len(datum.python_value),
                         1,
                         'Encoded value should be list of length 1')
        self.assertEqual(datum.python_value[0]['id'],
                         '1',
                         'id should be 1')
        self.assertEqual(datum.python_value[0]['name'],
                         'test',
                         'name should be test')

    def test_mysql_encode_mysqlrowlist(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "json")
        mysqlRowList: func.MySqlRowList = mysql.MySqlConverter.decode(
            data=datum, trigger_metadata=None)
        datum = mysql.MySqlConverter.encode(
            obj=mysqlRowList, expected_type=None)
        self.assertEqual(datum.type,
                         'json',
                         'Datum type should be JSON')
        self.assertEqual(len(datum.python_value),
                         1,
                         'Encoded value should be list of length 1')
        self.assertEqual(datum.python_value[0]['id'],
                         '1',
                         'id should be 1')
        self.assertEqual(datum.python_value[0]['name'],
                         'test',
                         'name should be test')

    def test_mysql_encode_list_of_mysqlrows(self):
        mysqlRows = [
            func.MySqlRow.from_json("""
        {
            "id": "1",
            "name": "test"
        }
        """),
            func.MySqlRow.from_json("""
            {
                "id": "2",
                "name": "test2"
            }
            """)
        ]
        datum = mysql.MySqlConverter.encode(obj=mysqlRows, expected_type=None)
        self.assertEqual(datum.type,
                         'json',
                         'Datum type should be JSON')
        self.assertEqual(len(datum.python_value),
                         2,
                         'Encoded value should be list of length 2')
        self.assertEqual(datum.python_value[0]['id'],
                         '1',
                         'id should be 1')
        self.assertEqual(datum.python_value[0]['name'],
                         'test',
                         'name should be test')
        self.assertEqual(datum.python_value[1]['id'],
                         '2',
                         'id should be 2')
        self.assertEqual(datum.python_value[1]['name'],
                         'test2',
                         'name should be test2')

    def test_mysql_encode_list_of_str_raises(self):
        strList = [
            """
        {
            "id": "1",
            "name": "test"
        }
        """
        ]
        self.assertRaises(NotImplementedError,
                          mysql.MySqlConverter.encode,
                          obj=strList,
                          expected_type=None)

    def test_mysql_encode_list_of_mysqlrowlist_raises(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "json")
        mysqlRowListList = [
            mysql.MySqlConverter.decode(
                data=datum, trigger_metadata=None)
        ]
        self.assertRaises(NotImplementedError,
                          mysql.MySqlConverter.encode,
                          obj=mysqlRowListList,
                          expected_type=None)

    def test_mysql_input_type(self):
        check_input_type = mysql.MySqlConverter.check_input_type_annotation
        self.assertTrue(check_input_type(func.MySqlRowList),
                        'MySqlRowList should be accepted')
        self.assertFalse(check_input_type(func.MySqlRow),
                         'MySqlRow should not be accepted')
        self.assertFalse(check_input_type(str),
                         'str should not be accepted')

    def test_mysql_output_type(self):
        check_output_type = mysql.MySqlConverter.check_output_type_annotation
        self.assertTrue(check_output_type(func.MySqlRowList),
                        'MySqlRowList should be accepted')
        self.assertTrue(check_output_type(func.MySqlRow),
                        'MySqlRow should be accepted')
        self.assertFalse(check_output_type(str),
                         'str should not be accepted')

    def test_mysqlrow_json(self):
        # Parse MySqlRow from JSON
        mysqlRow = func.MySqlRow.from_json("""
        {
            "id": "1",
            "name": "test"
        }
        """)
        self.assertEqual(mysqlRow['id'],
                         '1',
                         'Parsed MySqlRow id should be 1')
        self.assertEqual(mysqlRow['name'],
                         'test',
                         'Parsed MySqlRow name should be test')

        # Parse JSON from MySqlRow
        mysqlRowJson = json.loads(func.MySqlRow.to_json(mysqlRow))
        self.assertEqual(mysqlRowJson['id'],
                         '1',
                         'Parsed JSON id should be 1')
        self.assertEqual(mysqlRowJson['name'],
                         'test',
                         'Parsed JSON name should be test')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from abc import ABC, abstractmethod
from typing import Any, Union
from types import SimpleNamespace


"""
Azure Functions JSON utilities.
This module provides a JSON interface that can be used to serialize and
deserialize objects to and from JSON format. It supports both the `orjson`
and the standard `json` libraries, falling back to the standard library
if `orjson` is not available (installed).
"""


try:
    import orjson as _orjson
except ImportError:
    _orjson = None

# Standard library is always present
import json as _std_json


class JsonInterface(ABC):
    @abstractmethod
    def dumps(self, obj: Any) -> str:
        pass

    @abstractmethod
    def loads(self, s: Union[str, bytes, bytearray]) -> Any:
        pass


class OrJsonAdapter(JsonInterface):
    def __init__(self):
        assert _orjson is not None
        self.orjson = _orjson

    def dumps(self, obj: Any) -> str:
        # orjson.dumps returns bytes, decode to str
        return self.orjson.dumps(obj).decode("utf-8")

    def loads(self, s: Union[str, bytes, bytearray]) -> Any:
        return self.orjson.loads(s)


class StdJsonAdapter(JsonInterface):
    def __init__(self):
        self.json = _std_json

    def dumps(self, obj: Any) -> str:
        return self.json.dumps(obj)

    def loads(self, s: Union[str, bytes, bytearray]) -> Any:
        return self.json.loads(s)


if _orjson is not None:
    json_impl = OrJsonAdapter()
else:
    json_impl = StdJsonAdapter()


def dumps(obj: Any) -> str:
    return json_impl.dumps(obj)


def loads(s: Union[str, bytes, bytearray]) -> Any:
    return json_impl.loads(s)


json = SimpleNamespace(
    dumps=dumps,
    loads=loads
)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import TABLE
from azure.functions.decorators.core import DataType, OutputBinding, \
    InputBinding


class TableInput(InputBinding):

    @staticmethod
    def get_binding_name() -> str:
        return TABLE

    def __init__(self,
                 name: str,
                 connection: str,
                 table_name: str,
                 row_key: Optional[str] = None,
                 partition_key: Optional[str] = None,
                 take: Optional[int] = None,
                 filter: Optional[str] = None,
                 data_type: Optional[DataType] = None):
        self.connection = connection
        self.table_name = table_name
        self.row_key = row_key
        self.partition_key = partition_key
        self.take = take
        self.filter = filter
        super().__init__(name=name, data_type=data_type)


class TableOutput(OutputBinding):

    @staticmethod
    def get_binding_name() -> str:
        return TABLE

    def __init__(self,
                 name: str,
                 connection: str,
                 table_name: str,
                 row_key: Optional[str] = None,
                 partition_key: Optional[str] = None,
                 data_type: Optional[DataType] = None):
        self.connection = connection
        self.table_name = table_name
        self.row_key = row_key
        self.partition_key = partition_key
        super().__init__(name=name, data_type=data_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import List
import unittest
import json

from unittest.mock import patch
import azure.functions as func
import azure.functions.kafka as azf_ka
import azure.functions.meta as meta

from tests.utils.testutils import (CollectionBytes, CollectionString,
                                   CollectionSint64)


class Kafka(unittest.TestCase):
    SINGLE_KAFKA_DATAUM = '{"Offset":1,"Partition":0,"Topic":"users",'\
        '"Timestamp":"2020-06-20T04:43:28.998Z","Value":"hello", ' \
                          '"Headers":[{"Key":"test","Value":"1"}]}'
    SINGLE_KAFKA_TIMESTAMP = "2020-06-20T04:43:28.998Z"
    MULTIPLE_KAFKA_TIMESTAMP_0 = "2020-06-20T05:06:25.139Z"
    MULTIPLE_KAFKA_TIMESTAMP_1 = "2020-06-20T05:06:25.945Z"
    MULTIPLE_KAFKA_DATA_0 = '{"Offset":62,"Partition":1,"Topic":"message",'\
        '"Timestamp":"2020-06-20T05:06:25.139Z","Value":"a", ' \
                            '"Headers":[{"Key":"test","Value":"1"}], ' \
                            '"Key" : "1"}'
    MULTIPLE_KAFKA_DATA_1 = '{"Offset":63,"Partition":1,"Topic":"message",'\
        '"Timestamp":"2020-06-20T05:06:25.945Z","Value":"a", ' \
                            '"Headers":[{"Key":"test2","Value":"2"}], ' \
                            '"Key": "2"}'

    def test_kafka_input_type(self):
        check_input_type = (
            azf_ka.KafkaConverter.check_input_type_annotation
        )
        self.assertTrue(check_input_type(func.KafkaEvent))
        self.assertTrue(check_input_type(List[func.KafkaEvent]))
        self.assertFalse(check_input_type(str))
        self.assertFalse(check_input_type(bytes))
        self.assertFalse(check_input_type(List[str]))

    def test_kafka_output_type(self):
        check_output_type = (
            azf_ka.KafkaTriggerConverter.check_output_type_annotation
        )
        self.assertTrue(check_output_type(bytes))
        self.assertTrue(check_output_type(str))
        self.assertTrue(check_output_type(List[str]))
        self.assertFalse(check_output_type(func.KafkaEvent))
        self.assertFalse(check_output_type(List[bytes]))
        self.assertFalse(check_output_type(List[func.KafkaEvent]))

    @patch('azure.functions.kafka.KafkaTriggerConverter'
           '.decode_single_event')
    @patch('azure.functions.kafka.KafkaTriggerConverter'
           '.decode_multiple_events')
    def test_kafka_decode_single_event(self, dme_mock, dse_mock):
        azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_single_kafka_datum(),
            trigger_metadata=self._generate_single_trigger_metadatum()
        )
        dse_mock.assert_called_once()
        dme_mock.assert_not_called()

    @patch('azure.functions.kafka.KafkaTriggerConverter'
           '.decode_single_event')
    @patch('azure.functions.kafka.KafkaTriggerConverter'
           '.decode_multiple_events')
    def test_kafka_decode_multiple_events(self, dse_mock, dme_mock):
        azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_multiple_kafka_data(),
            trigger_metadata=self._generate_multiple_trigger_metadata()
        )
        dse_mock.assert_not_called()
        dme_mock.assert_called_once()

    def test_kafka_trigger_single_event_json(self):
        result = azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_single_kafka_datum('json'),
            trigger_metadata=self._generate_single_trigger_metadatum()
        )
        # Result body always has the datatype of bytes
        self.assertEqual(
            result.get_body().decode('utf-8'), self.SINGLE_KAFKA_DATAUM
        )
        self.assertEqual(result.timestamp, self.SINGLE_KAFKA_TIMESTAMP)

    def test_kafka_trigger_single_event_bytes(self):
        result = azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_single_kafka_datum('bytes'),
            trigger_metadata=self._generate_single_trigger_metadatum()
        )
        self.assertEqual(
            result.get_body().decode('utf-8'), self.SINGLE_KAFKA_DATAUM
        )
        self.assertEqual(result.timestamp, self.SINGLE_KAFKA_TIMESTAMP)

    def test_kafka_trigger_single_event_string(self):
        result = azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_single_kafka_datum('string'),
            trigger_metadata=self._generate_single_trigger_metadatum()
        )
        self.assertEqual(
            result.get_body().decode('utf-8'), self.SINGLE_KAFKA_DATAUM
        )
        self.assertEqual(result.timestamp, self.SINGLE_KAFKA_TIMESTAMP)

    def test_kafka_trigger_multiple_events_collection_string(self):
        result = azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_multiple_kafka_data('collection_string'),
            trigger_metadata=self._generate_multiple_trigger_metadata()
        )
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)

        self.assertEqual(
            result[0].timestamp,
            self.MULTIPLE_KAFKA_TIMESTAMP_0)
        self.assertEqual(
            json.loads(result[0].get_body().decode('utf-8')),
            json.loads(self.MULTIPLE_KAFKA_DATA_0)
        )

        self.assertEqual(
            result[1].timestamp,
            self.MULTIPLE_KAFKA_TIMESTAMP_1)
        self.assertEqual(
            json.loads(result[1].get_body().decode('utf-8')),
            json.loads(self.MULTIPLE_KAFKA_DATA_1)
        )

    def test_kafka_trigger_multiple_events_collection_bytes(self):
        result = azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_multiple_kafka_data('collection_bytes'),
            trigger_metadata=self._generate_multiple_trigger_metadata()
        )
        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)

        self.assertEqual(
            result[0].timestamp,
            self.MULTIPLE_KAFKA_TIMESTAMP_0)
        self.assertEqual(
            json.loads(result[0].get_body().decode('utf-8')),
            json.loads(self.MULTIPLE_KAFKA_DATA_0)
        )

        self.assertEqual(
            result[1].timestamp,
            self.MULTIPLE_KAFKA_TIMESTAMP_1)
        self.assertEqual(
            json.loads(result[1].get_body().decode('utf-8')),
            json.loads(self.MULTIPLE_KAFKA_DATA_1)
        )

    def test_single_kafka_trigger_metadata_field(self):
        result = azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_single_kafka_datum(),
            trigger_metadata=self._generate_single_trigger_metadatum()
        )

        # Timestamp
        self.assertEqual(result.timestamp, self.SINGLE_KAFKA_TIMESTAMP)
        # Offset should be 1
        self.assertEqual(result.offset, 1)
        # Topic
        self.assertEqual(result.topic, "users")
        # Partition
        self.assertEqual(result.partition, 0)
        # Value
        self.assertEqual(
            json.loads(result.get_body().decode('utf-8'))['Value'], "hello")
        # Metadata
        metadata_dict = result.metadata
        sys = metadata_dict['sys']
        sys_dict = json.loads(sys)
        self.assertEqual(sys_dict['MethodName'], 'KafkaTrigger')

    def test_multiple_kafka_triggers_metadata_field(self):
        result = azf_ka.KafkaTriggerConverter.decode(
            data=self._generate_multiple_kafka_data("collection_string"),
            trigger_metadata=self._generate_multiple_trigger_metadata()
        )

        self.assertEqual(result[0].offset, 62)
        self.assertEqual(result[1].offset, 63)

        self.assertEqual(result[0].partition, 1)
        self.assertEqual(result[1].partition, 2)

        self.assertEqual(
            result[0].timestamp,
            "2020-06-20T05:06:25.139Z")
        self.assertEqual(
            result[1].timestamp,
            "2020-06-20T05:06:25.945Z")
        metadata_dict = result[0].metadata
        sys = metadata_dict['sys']
        sys_dict = json.loads(sys)
        self.assertEqual(sys_dict['MethodName'], 'KafkaTriggerMany')

    def test_kafka_convert_single_event_str(self):
        datum = meta.Datum("dummy_body", "string")
        result = azf_ka.KafkaConverter.decode(
            data=datum, trigger_metadata=None)

        self.assertEqual(result.get_body(), b"dummy_body")

    def test_kafka_convert_single_event_bytes(self):
        datum = meta.Datum(b"dummy_body", "bytes")
        result = azf_ka.KafkaConverter.decode(
            data=datum, trigger_metadata=None)

        self.assertEqual(result.get_body(), b"dummy_body")

    def test_kafka_convert_multiple_event_collection_string(self):
        datum = meta.Datum(CollectionString(["dummy_body1", "dummy_body2"]),
                           "collection_string")
        result = azf_ka.KafkaConverter.decode(
            data=datum, trigger_metadata=None)

        self.assertEqual([result[0].get_body(), result[1].get_body()],
                         [b"dummy_body1", b"dummy_body2"])

    def test_kafka_convert_multiple_event_collection_bytes(self):
        datum = meta.Datum(CollectionBytes([b"dummy_body1", b"dummy_body2"]),
                           "collection_bytes")
        result = azf_ka.KafkaConverter.decode(
            data=datum, trigger_metadata=None)

        self.assertEqual([result[0].get_body(), result[1].get_body()],
                         [b"dummy_body1", b"dummy_body2"])

    def _generate_single_kafka_datum(self, datum_type='string'):
        datum = self.SINGLE_KAFKA_DATAUM
        if datum_type == 'bytes':
            datum = datum.encode('utf-8')
        return meta.Datum(datum, datum_type)

    def _generate_multiple_kafka_data(self, data_type='json'):
        data = '[{"Offset":62,"Partition":1,"Topic":"message",'\
               '"Timestamp":"2020-06-20T05:06:25.139Z","Value":"a",' \
               ' "Headers":[{"Key":"test","Value":"1"}], "Key": "1"},'\
               ' {"Offset":63,"Partition":1,"Topic":"message",'\
               '"Timestamp":"2020-06-20T05:06:25.945Z","Value":"a", ' \
               '"Headers":[{"Key":"test2","Value":"2"}], "Key": "2"}]'
        if data_type == 'collection_bytes':
            data = list(
                map(lambda x: json.dumps(x).encode('utf-8'),
                    json.loads(data))
            )
            data = CollectionBytes(data)
        elif data_type == 'collection_string':
            data = list(
                map(lambda x: json.dumps(x), json.loads(data))
            )
            data = CollectionString(data)

        return meta.Datum(data, data_type)

    def _generate_single_trigger_metadatum(self):
        return {
            'Offset': meta.Datum(
                '1', 'string'
            ),
            'Partition': meta.Datum(
                '0', 'string'
            ),
            'Timestamp': meta.Datum(
                self.SINGLE_KAFKA_TIMESTAMP, 'string'
            ),
            'Topic': meta.Datum(
                'users', 'string'
            ),
            'Value': meta.Datum(
                'hello', 'string'
            ),
            'sys': meta.Datum(
                '{"MethodName":"KafkaTrigger",'
                '"UtcNow":"2020-06-20T04:43:30.6756278Z",'
                '"RandGuid":"b0870c0c-2b7a-40dc-b4be-45224c91a49c"}',
                'json'
            )  # __len__: 6
        }

    def _generate_multiple_trigger_metadata(self):
        key_array = ["1", "2"]
        partition_array = [1, 2]
        timestamp_array = ["2020-06-20T05:06:25.139Z",
                           "2020-06-20T05:06:25.945Z"]

        return {
            'KeyArray': meta.Datum(
                json.dumps(key_array), 'json'
            ),
            'OffsetArray': meta.Datum(
                CollectionSint64([62, 63]), 'collection_sint64'
            ),
            'PartitionArray': meta.Datum(
                json.dumps(partition_array), 'json'
            ),
            'TimestampArray': meta.Datum(
                json.dumps(timestamp_array), 'json'
            ),
            'TopicArray': meta.Datum(
                CollectionString(['message', 'message']), "collection_string"
            ),
            'HeadersArray': meta.Datum(
                json.dumps([['{"Key":"test","Value":"1"}'],
                            ['{"Key":"test2","Value":"2"}']]), 'json'
            ),
            'sys': meta.Datum(
                '{"MethodName":"KafkaTriggerMany",'
                '"UtcNow":"2020-06-20T05:06:26.5550868Z",'
                '"RandGuid":"57d5eeb7-c86c-4924-a14a-160092154093"}',
                'json'
            )
        }



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import KAFKA, KAFKA_TRIGGER
from azure.functions.decorators.core import Cardinality, DataType, \
    OutputBinding, Trigger
from .utils import StringifyEnum


class BrokerAuthenticationMode(StringifyEnum):
    NOTSET = -1
    GSSAPI = 0
    PLAIN = 1
    SCRAMSHA256 = 2
    SCRAMSHA512 = 3


class BrokerProtocol(StringifyEnum):
    NOTSET = -1
    PLAINTEXT = 0
    SSL = 1
    SASLPLAINTEXT = 2
    SASLSSL = 3


class OAuthBearerMethod(StringifyEnum):
    DEFAULT = 0,
    OIDC = 1


class KafkaOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return KAFKA

    def __init__(self,
                 name: str,
                 topic: str,
                 broker_list: str,
                 avro_schema: Optional[str],
                 username: Optional[str],
                 password: Optional[str],
                 ssl_key_location: Optional[str],
                 ssl_ca_location: Optional[str],
                 ssl_certificate_location: Optional[str],
                 ssl_key_password: Optional[str],
                 schema_registry_url: Optional[str],
                 schema_registry_username: Optional[str],
                 schema_registry_password: Optional[str],
                 o_auth_bearer_method: Optional[OAuthBearerMethod] = None,
                 o_auth_bearer_client_id: Optional[str] = None,
                 o_auth_bearer_client_secret: Optional[str] = None,
                 o_auth_bearer_scope: Optional[str] = None,
                 o_auth_bearer_token_endpoint_url: Optional[str] = None,
                 o_auth_bearer_extensions: Optional[str] = None,
                 max_message_bytes: int = 1_000_000,
                 batch_size: int = 10_000,
                 enable_idempotence: bool = False,
                 message_timeout_ms: int = 300_000,
                 request_timeout_ms: int = 5_000,
                 max_retries: int = 2_147_483_647,
                 authentication_mode: Optional[BrokerAuthenticationMode] = BrokerAuthenticationMode.NOTSET,  # noqa: E501
                 protocol: Optional[BrokerProtocol] = BrokerProtocol.NOTSET,
                 linger_ms: int = 5,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.topic = topic
        self.broker_list = broker_list
        self.avro_schema = avro_schema
        self.username = username
        self.password = password
        self.ssl_key_location = ssl_key_location
        self.ssl_ca_location = ssl_ca_location
        self.ssl_certificate_location = ssl_certificate_location
        self.ssl_key_password = ssl_key_password
        self.schema_registry_url = schema_registry_url
        self.schema_registry_username = schema_registry_username
        self.schema_registry_password = schema_registry_password
        self.o_auth_bearer_method = o_auth_bearer_method
        self.o_auth_bearer_client_id = o_auth_bearer_client_id
        self.o_auth_bearer_client_secret = o_auth_bearer_client_secret
        self.o_auth_bearer_scope = o_auth_bearer_scope
        self.o_auth_bearer_token_endpoint_url = o_auth_bearer_token_endpoint_url  # noqa: E501
        self.o_auth_bearer_extensions = o_auth_bearer_extensions
        self.max_message_bytes = max_message_bytes
        self.batch_size = batch_size
        self.enable_idempotence = enable_idempotence
        self.message_timeout_ms = message_timeout_ms
        self.request_timeout_ms = request_timeout_ms
        self.max_retries = max_retries
        self.authentication_mode = authentication_mode
        self.protocol = protocol
        self.linger_ms = linger_ms
        super().__init__(name=name, data_type=data_type)


class KafkaTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return KAFKA_TRIGGER

    def __init__(self,
                 name: str,
                 topic: str,
                 broker_list: str,
                 event_hub_connection_string: Optional[str],
                 consumer_group: Optional[str],
                 avro_schema: Optional[str],
                 username: Optional[str],
                 password: Optional[str],
                 ssl_key_location: Optional[str],
                 ssl_ca_location: Optional[str],
                 ssl_certificate_location: Optional[str],
                 ssl_key_password: Optional[str],
                 schema_registry_url: Optional[str],
                 schema_registry_username: Optional[str],
                 schema_registry_password: Optional[str],
                 o_auth_bearer_method: Optional[OAuthBearerMethod] = None,
                 o_auth_bearer_client_id: Optional[str] = None,
                 o_auth_bearer_client_secret: Optional[str] = None,
                 o_auth_bearer_scope: Optional[str] = None,
                 o_auth_bearer_token_endpoint_url: Optional[str] = None,
                 o_auth_bearer_extensions: Optional[str] = None,
                 authentication_mode: Optional[BrokerAuthenticationMode] = BrokerAuthenticationMode.NOTSET,  # noqa: E501
                 protocol: Optional[BrokerProtocol] = BrokerProtocol.NOTSET,
                 cardinality: Optional[Cardinality] = Cardinality.ONE,
                 lag_threshold: int = 1000,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.topic = topic
        self.broker_list = broker_list
        self.event_hub_connection_string = event_hub_connection_string
        self.consumer_group = consumer_group
        self.avro_schema = avro_schema
        self.username = username
        self.password = password
        self.ssl_key_location = ssl_key_location
        self.ssl_ca_location = ssl_ca_location
        self.ssl_certificate_location = ssl_certificate_location
        self.ssl_key_password = ssl_key_password
        self.schema_registry_url = schema_registry_url
        self.schema_registry_username = schema_registry_username
        self.schema_registry_password = schema_registry_password
        self.o_auth_bearer_method = o_auth_bearer_method
        self.o_auth_bearer_client_id = o_auth_bearer_client_id
        self.o_auth_bearer_client_secret = o_auth_bearer_client_secret
        self.o_auth_bearer_scope = o_auth_bearer_scope
        self.o_auth_bearer_token_endpoint_url = o_auth_bearer_token_endpoint_url  # noqa: E501
        self.o_auth_bearer_extensions = o_auth_bearer_extensions
        self.authentication_mode = authentication_mode
        self.protocol = protocol
        self.cardinality = cardinality
        self.lag_threshold = lag_threshold
        super().__init__(name=name, data_type=data_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing

from . import meta


class WarmUpContext:
    pass


class WarmUpTriggerConverter(meta.InConverter, binding='warmupTrigger',
                             trigger=True):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, WarmUpContext)

    @classmethod
    def decode(cls, data: meta.Datum, *, trigger_metadata) -> typing.Any:
        return WarmUpContext()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json
import unittest

import azure.functions.timer as timer
from azure.functions.meta import Datum


class TestTimer(unittest.TestCase):
    def test_timer_decode(self):
        # given
        data = '''{"Schedule":{"AdjustForDST":true},
                   "ScheduleStatus":{
                        "Last":"2022-03-28T15:40:00.0105419-05:00",
                        "Next":"2022-03-28T15:45:00-05:00",
                        "LastUpdated":"2022-03-28T15:40:00.0105419-05:00"},
                   "IsPastDue":false}'''
        datum: Datum = Datum(value=data, type='json')
        data_dict = json.loads(data)

        # when
        timer_request: timer.TimerRequest = \
            timer.TimerRequestConverter.decode(datum, trigger_metadata={})

        # then
        self.assertEqual(timer_request.schedule, data_dict["Schedule"])
        self.assertEqual(timer_request.schedule_status,
                         data_dict["ScheduleStatus"])
        self.assertEqual(timer_request.past_due, data_dict["IsPastDue"])

    def test_timer_initialize_without_args(self):
        # given
        past_due = False
        schedule_status = {}
        schedule = {}

        # when
        test_timer = timer.TimerRequest()

        # then
        self.assertEqual(past_due, test_timer.past_due)
        self.assertEqual(schedule_status, test_timer.schedule_status)
        self.assertEqual(schedule, test_timer.schedule)

    def test_timer_initialize_empty_dicts(self):
        # given
        past_due = False

        # when
        test_timer = timer.TimerRequest()

        # then
        self.assertEqual(past_due, test_timer.past_due)
        self.assertEqual({}, test_timer.schedule_status)
        self.assertEqual({}, test_timer.schedule)

    def test_timer_no_implementation_exception(self):
        # given
        datum: Datum = Datum(value="test", type='string')
        is_exception_raised = False

        # when
        try:
            timer.TimerRequestConverter.decode(datum, trigger_metadata={})
        except NotImplementedError:
            is_exception_raised = True

        # then
        self.assertTrue(is_exception_raised)

    def test_timer_input_type(self):
        check_input_type = (
            timer.TimerRequestConverter.check_input_type_annotation
        )
        self.assertTrue(check_input_type(timer.TimerRequest))
        self.assertFalse(check_input_type(str))



from typing import Optional, Union

from azure.functions.decorators.constants import (ASSISTANT_SKILL_TRIGGER,
                                                  TEXT_COMPLETION,
                                                  ASSISTANT_QUERY,
                                                  EMBEDDINGS, EMBEDDINGS_STORE,
                                                  ASSISTANT_CREATE,
                                                  ASSISTANT_POST,
                                                  SEMANTIC_SEARCH)
from azure.functions.decorators.core import Trigger, DataType, InputBinding, \
    OutputBinding
from azure.functions.decorators.utils import StringifyEnum


class InputType(StringifyEnum):

    RawText = "raw_text",
    FilePath = "file_path"


class OpenAIModels(StringifyEnum):
    DefaultChatModel = "gpt-3.5-turbo"
    DefaultEmbeddingsModel = "text-embedding-ada-002"


class AssistantSkillTrigger(Trigger):

    @staticmethod
    def get_binding_name() -> str:
        return ASSISTANT_SKILL_TRIGGER

    def __init__(self,
                 name: str,
                 function_description: str,
                 function_name: Optional[str] = None,
                 parameter_description_json: Optional[str] = None,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.function_description = function_description
        self.function_name = function_name
        self.parameter_description_json = parameter_description_json
        super().__init__(name=name, data_type=data_type)


class TextCompletionInput(InputBinding):

    @staticmethod
    def get_binding_name() -> str:
        return TEXT_COMPLETION

    def __init__(self,
                 name: str,
                 prompt: str,
                 ai_connection_name: Optional[str] = "",
                 chat_model: Optional
                 [Union[str, OpenAIModels]]
                 = OpenAIModels.DefaultChatModel,
                 temperature: Optional[str] = "0.5",
                 top_p: Optional[str] = None,
                 max_tokens: Optional[str] = "100",
                 is_reasoning_model: Optional[bool] = False,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.prompt = prompt
        self.ai_connection_name = ai_connection_name
        self.chat_model = chat_model
        self.temperature = temperature
        self.top_p = top_p
        self.max_tokens = max_tokens
        self.is_reasoning_model = is_reasoning_model
        super().__init__(name=name, data_type=data_type)


class AssistantQueryInput(InputBinding):

    @staticmethod
    def get_binding_name():
        return ASSISTANT_QUERY

    def __init__(self,
                 name: str,
                 id: str,
                 timestamp_utc: str,
                 chat_storage_connection_setting: Optional[str] = "AzureWebJobsStorage",        # noqa: E501
                 collection_name: Optional[str] = "ChatState",
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.id = id
        self.timestamp_utc = timestamp_utc
        self.chat_storage_connection_setting = chat_storage_connection_setting
        self.collection_name = collection_name
        super().__init__(name=name, data_type=data_type)


class EmbeddingsInput(InputBinding):

    @staticmethod
    def get_binding_name() -> str:
        return EMBEDDINGS

    def __init__(self,
                 name: str,
                 input: str,
                 input_type: InputType,
                 ai_connection_name: Optional[str] = "",
                 embeddings_model: Optional
                 [Union[str, OpenAIModels]]
                 = OpenAIModels.DefaultEmbeddingsModel,
                 max_chunk_length: Optional[int] = 8 * 1024,
                 max_overlap: Optional[int] = 128,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.name = name
        self.input = input
        self.input_type = input_type
        self.ai_connection_name = ai_connection_name
        self.embeddings_model = embeddings_model
        self.max_chunk_length = max_chunk_length
        self.max_overlap = max_overlap
        super().__init__(name=name, data_type=data_type)


semantic_search_system_prompt = \
    """You are a helpful assistant. You are responding to requests
    from a user about internal emails and documents. You can and
    should refer to the internal documents to help respond to
    requests. If a user makes a request that's not covered by the
    internal emails and documents, explain that you don't know the
    answer or that you don't have access to the information.

    The following is a list of documents that you can refer to when
    answering questions. The documents are in the format
    [filename]: [text] and are separated by newlines. If you answer
    a question by referencing any of the documents, please cite the
    document in your answer. For example, if you answer a question
    by referencing info.txt, you should add "Reference: info.txt"
    to the end of your answer on a separate line."""


class SemanticSearchInput(InputBinding):

    @staticmethod
    def get_binding_name() -> str:
        return SEMANTIC_SEARCH

    def __init__(self,
                 name: str,
                 search_connection_name: str,
                 collection: str,
                 query: Optional[str] = None,
                 ai_connection_name: Optional[str] = "",
                 embeddings_model: Optional
                 [Union[str, OpenAIModels]]
                 = OpenAIModels.DefaultEmbeddingsModel,
                 chat_model: Optional
                 [Union[str, OpenAIModels]]
                 = OpenAIModels.DefaultChatModel,
                 system_prompt: Optional[str] = semantic_search_system_prompt,
                 max_knowledge_count: Optional[int] = 1,
                 temperature: Optional[str] = "0.5",
                 top_p: Optional[str] = None,
                 max_tokens: Optional[str] = "100",
                 is_reasoning_model: Optional[bool] = False,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.name = name
        self.search_connection_name = search_connection_name
        self.collection = collection
        self.query = query
        self.ai_connection_name = ai_connection_name
        self.embeddings_model = embeddings_model
        self.chat_model = chat_model
        self.system_prompt = system_prompt
        self.max_knowledge_count = max_knowledge_count
        self.temperature = temperature
        self.top_p = top_p
        self.max_tokens = max_tokens
        self.is_reasoning_model = is_reasoning_model
        super().__init__(name=name, data_type=data_type)


class AssistantPostInput(InputBinding):

    @staticmethod
    def get_binding_name():
        return ASSISTANT_POST

    def __init__(self, name: str,
                 id: str,
                 user_message: str,
                 ai_connection_name: Optional[str] = "",
                 chat_model: Optional
                 [Union[str, OpenAIModels]]
                 = OpenAIModels.DefaultChatModel,
                 chat_storage_connection_setting: Optional[str] = "AzureWebJobsStorage",       # noqa: E501
                 collection_name: Optional[str] = "ChatState",
                 temperature: Optional[str] = "0.5",
                 top_p: Optional[str] = None,
                 max_tokens: Optional[str] = "100",
                 is_reasoning_model: Optional[bool] = False,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.name = name
        self.id = id
        self.user_message = user_message
        self.ai_connection_name = ai_connection_name
        self.chat_model = chat_model
        self.chat_storage_connection_setting = chat_storage_connection_setting
        self.collection_name = collection_name
        self.temperature = temperature
        self.top_p = top_p
        self.max_tokens = max_tokens
        self.is_reasoning_model = is_reasoning_model
        super().__init__(name=name, data_type=data_type)


class EmbeddingsStoreOutput(OutputBinding):

    @staticmethod
    def get_binding_name() -> str:
        return EMBEDDINGS_STORE

    def __init__(self,
                 name: str,
                 input: str,
                 input_type: InputType,
                 store_connection_name: str,
                 collection: str,
                 ai_connection_name: Optional[str] = "",
                 embeddings_model: Optional
                 [Union[str, OpenAIModels]]
                 = OpenAIModels.DefaultEmbeddingsModel,
                 max_chunk_length: Optional[int] = 8 * 1024,
                 max_overlap: Optional[int] = 128,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.name = name
        self.input = input
        self.input_type = input_type
        self.store_connection_name = store_connection_name
        self.collection = collection
        self.ai_connection_name = ai_connection_name
        self.embeddings_model = embeddings_model
        self.max_chunk_length = max_chunk_length
        self.max_overlap = max_overlap
        super().__init__(name=name, data_type=data_type)


class AssistantCreateOutput(OutputBinding):

    @staticmethod
    def get_binding_name():
        return ASSISTANT_CREATE

    def __init__(self,
                 name: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        super().__init__(name=name, data_type=data_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import logging
from io import BytesIO, StringIO
from os import linesep
from typing import Dict, List, Optional, Any
from urllib.parse import ParseResult, urlparse, unquote_to_bytes
from wsgiref.headers import Headers

from ._abc import Context
from ._http import HttpRequest, HttpResponse


def wsgi_encoding_dance(value):
    return value.encode().decode("latin1")


class WsgiRequest:
    _environ_cache: Optional[Dict[str, Any]] = None

    def __init__(self,
                 func_req: HttpRequest,
                 func_ctx: Optional[Context] = None):
        url: ParseResult = urlparse(func_req.url)
        func_req_body = func_req.get_body() or b''

        # Convert function request headers to lowercase header
        self._lowercased_headers = {
            k.lower(): v for k, v in func_req.headers.items()
        }

        # Implement interfaces for PEP 3333 environ
        self.request_method = getattr(func_req, 'method', None)
        self.script_name = ''
        self.path_info = (
            unquote_to_bytes(getattr(url, 'path', None))  # type: ignore
            .decode('latin-1' if type(self) is WsgiRequest else 'utf-8')
        )

        self.query_string = getattr(url, 'query', None)
        self.content_type = self._lowercased_headers.get('content-type')
        self.content_length = str(len(func_req_body))
        self.server_name = getattr(url, 'hostname', None)
        self.server_port = str(self._get_port(url, self._lowercased_headers))
        self.server_protocol = 'HTTP/1.1'

        # Propagate http request headers into HTTP_ environ
        self._http_environ: Dict[str, str] = self._get_http_headers(
            func_req.headers
        )

        # Wsgi environ
        self.wsgi_version = (1, 0)
        self.wsgi_url_scheme = url.scheme
        self.wsgi_input = BytesIO(func_req_body)
        self.wsgi_multithread = False
        self.wsgi_multiprocess = False
        self.wsgi_run_once = False

        # Azure Functions context
        self.af_function_directory = getattr(func_ctx,
                                             'function_directory', None)
        self.af_function_name = getattr(func_ctx, 'function_name', None)
        self.af_invocation_id = getattr(func_ctx, 'invocation_id', None)
        self.af_thread_local_storage = getattr(func_ctx,
                                               'thread_local_storage',
                                               None)
        self.af_trace_context = getattr(func_ctx, 'trace_context', None)
        self.af_retry_context = getattr(func_ctx, 'retry_context', None)

    def to_environ(self, errors_buffer: StringIO) -> Dict[str, Any]:
        if self._environ_cache is not None:
            return self._environ_cache

        environ = {
            'REQUEST_METHOD': self.request_method,
            'SCRIPT_NAME': self.script_name,
            'PATH_INFO': self.path_info,
            'QUERY_STRING': self.query_string,
            'CONTENT_TYPE': self.content_type,
            'CONTENT_LENGTH': self.content_length,
            'SERVER_NAME': self.server_name,
            'SERVER_PORT': self.server_port,
            'SERVER_PROTOCOL': self.server_protocol,
            'wsgi.version': self.wsgi_version,
            'wsgi.url_scheme': self.wsgi_url_scheme,
            'wsgi.input': self.wsgi_input,
            'wsgi.errors': errors_buffer,
            'wsgi.multithread': self.wsgi_multithread,
            'wsgi.multiprocess': self.wsgi_multiprocess,
            'wsgi.run_once': self.wsgi_run_once,
            'azure_functions.function_directory': self.af_function_directory,
            'azure_functions.function_name': self.af_function_name,
            'azure_functions.invocation_id': self.af_invocation_id,
            'azure_functions.thread_local_storage':
                self.af_thread_local_storage,
            'azure_functions.trace_context': self.af_trace_context,
            'azure_functions.retry_context': self.af_retry_context
        }
        environ.update(self._http_environ)

        # Ensure WSGI string fits in IOS-8859-1 code points
        for k, v in environ.items():
            if isinstance(v, (str,)):
                environ[k] = wsgi_encoding_dance(v)

        # Remove None values
        self._environ_cache = {
            k: v for k, v in environ.items() if v is not None
        }
        return self._environ_cache

    def _get_port(self, parsed_url, lowercased_headers: Dict[str, str]) -> int:
        port: int = 80
        if lowercased_headers.get('x-forwarded-port'):
            return int(lowercased_headers['x-forwarded-port'])
        elif getattr(parsed_url, 'port', None):
            return int(parsed_url.port)
        elif parsed_url.scheme == 'https':
            return 443
        return port

    def _get_http_headers(self,
                          func_headers: Dict[str, str]) -> Dict[str, str]:
        # Content-Type -> HTTP_CONTENT_TYPE
        return {f'HTTP_{k.upper().replace("-", "_")}': v for k, v in
                func_headers.items()}


class WsgiResponse:
    def __init__(self):
        self._status = ''
        self._status_code = 0
        self._headers = {}
        self._buffer: List[bytes] = []

    @classmethod
    def from_app(cls, app, environ) -> 'WsgiResponse':
        res = cls()
        res._buffer = [x or b'' for x in app(environ, res._start_response)]
        return res

    def to_func_response(self) -> HttpResponse:
        lowercased_headers = {k.lower(): v for k, v in self._headers.items()}
        return HttpResponse(
            body=b''.join(self._buffer),
            status_code=self._status_code,
            headers=self._headers,
            mimetype=lowercased_headers.get('content-type'),
            charset=lowercased_headers.get('content-encoding')
        )

    # PEP 3333 start response implementation
    def _start_response(self, status: str, response_headers: List[Any]):
        self._status = status
        self._headers = Headers(response_headers)  # type: ignore
        self._status_code = int(self._status.split(' ')[0])  # 200 OK


class WsgiMiddleware:
    """This middleware is to adapt a WSGI supported Python server
    framework into Azure Functions. It can be used by either calling the
    .handle() function or exposing the .main property in a HttpTrigger.
    """
    _logger = logging.getLogger('azure.functions.WsgiMiddleware')
    _usage_reported = False

    def __init__(self, app):
        """Instantiate a WSGI middleware to convert Azure Functions HTTP
        request into WSGI Python object. Example on handling WSGI app in a HTTP
        trigger by overwriting the .main() method:

        import azure.functions as func

        from FlaskApp import app

        main = func.WsgiMiddleware(app.wsgi_app).main
        """
        if not self._usage_reported:
            self._logger.info("Instantiating Azure Functions WSGI middleware.")
            self._usage_reported = True

        self._app = app
        self._wsgi_error_buffer = StringIO()
        self.main = self._handle

    def handle(self, req: HttpRequest, context: Optional[Context] = None):
        """Method to convert an Azure Functions HTTP request into a WSGI
        Python object. Example on handling WSGI app in a HTTP trigger by
        calling .handle() in .main() method:

        import azure.functions as func

        from FlaskApp import app

        def main(req, context):
            return func.WsgiMiddleware(app.wsgi_app).handle(req, context)
        """
        return self._handle(req, context)

    def _handle(self, req, context):
        wsgi_request = WsgiRequest(req, context)
        environ = wsgi_request.to_environ(self._wsgi_error_buffer)
        wsgi_response = WsgiResponse.from_app(self._app, environ)
        self._handle_errors(wsgi_response)
        return wsgi_response.to_func_response()

    def _handle_errors(self, wsgi_response):
        if self._wsgi_error_buffer.tell() > 0:
            self._wsgi_error_buffer.seek(0)
            error_message = linesep.join(
                self._wsgi_error_buffer.readline()
            )
            raise Exception(error_message)

        if wsgi_response._status_code >= 500:
            raise Exception(b''.join(wsgi_response._buffer))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from abc import ABC, abstractmethod
from typing import Dict, Optional, Type

from .utils import to_camel_case, \
    ABCBuildDictMeta, StringifyEnum

SCRIPT_FILE_NAME = "function_app.py"


# Enums
class BindingDirection(StringifyEnum):
    """Direction of the binding used in function.json"""
    IN = 0
    """Input binding direction."""
    OUT = 1
    """Output binding direction."""
    INOUT = 2
    """Some bindings support a special binding direction. """


class DataType(StringifyEnum):
    """Data type of the binding used in function.json"""
    """Parse binding argument as undefined."""
    UNDEFINED = 0
    """Parse binding argument as string."""
    STRING = 1
    """Parse binding argument as binary."""
    BINARY = 2
    """Parse binding argument as stream."""
    STREAM = 3


class AuthLevel(StringifyEnum):
    """Azure HTTP authorization level, determines what keys, if any, need to
    be present on the request in order to invoke the function. """
    FUNCTION = "function"
    """A function-specific API key is required. This is the default value if
    none is provided. """
    ANONYMOUS = "anonymous"
    """No API key is required."""
    ADMIN = "admin"
    """The master key is required."""


class Cardinality(StringifyEnum):
    """Used for all non-C# languages. Set to many in order to enable
    batching. If omitted or set to one, a single message is passed to the
    function. """
    ONE = "one"
    """Singe message passed to the function."""
    MANY = "many"
    """Multiple messaged passed to the function per invocation."""


class AccessRights(StringifyEnum):
    """Access rights for the connection string. The default is manage,
    which indicates that the connection has the Manage permission. """
    MANAGE = "manage"
    """Confers the right to manage the topology of the namespace, including
    creating and deleting entities. """
    LISTEN = "listen"
    """Confers the right to listen (relay) or receive (queue, subscriptions)
    and all related message handling. """


class BlobSource(StringifyEnum):
    """Source of the blob trigger."""
    EVENT_GRID = "EventGrid"
    """Event Grid is the source of the blob trigger."""
    LOGS_AND_CONTAINER_SCAN = "LogsAndContainerScan"
    """Standard polling mechanism to detect changes in the container."""


class Binding(ABC):
    """Abstract binding class which captures common attributes and
    functions. :meth:`get_dict_repr` can auto generate the function.json for
    every binding, the only restriction is ***ENSURE*** __init__ parameter
    names of any binding class are snake case form of corresponding
    attribute in function.json when new binding classes are created.
    Ref: https://aka.ms/azure-function-binding-http """

    EXCLUDED_INIT_PARAMS = {'self', 'kwargs', 'type', 'data_type', 'direction'}

    @staticmethod
    @abstractmethod
    def get_binding_name() -> str:
        pass

    def __init__(self, name: str,
                 direction: BindingDirection,
                 data_type: Optional[DataType] = None,
                 type: Optional[str] = None):  # NoQa
        # For natively supported bindings, get_binding_name is always
        # implemented, and for generic bindings, type is a required argument
        # in decorator functions.
        self.type = self.get_binding_name() \
            if self.get_binding_name() is not None else type
        self.name = name
        self._direction = direction
        self._data_type = data_type
        self._dict = {
            "direction": self._direction,
            "dataType": self._data_type,
            "type": self.type
        }

    @property
    def data_type(self) -> Optional[int]:
        return self._data_type.value if self._data_type else None

    @property
    def direction(self) -> int:
        return self._direction.value

    def get_dict_repr(self) -> Dict:
        """Build a dictionary of a particular binding. The keys are camel
        cased binding field names defined in `init_params` list and
        :class:`Binding` class. \n
        This method is invoked in function :meth:`get_raw_bindings` of class
        :class:`Function` to generate json dict for each binding.

        :return: Dictionary representation of the binding.
        """
        params = list(dict.fromkeys(getattr(self, 'init_params', [])))
        for p in params:
            if p not in Binding.EXCLUDED_INIT_PARAMS:
                self._dict[to_camel_case(p)] = getattr(self, p, None)

        return self._dict


class Trigger(Binding, ABC, metaclass=ABCBuildDictMeta):
    """Class representation of Azure Function Trigger. \n
    Ref: https://aka.ms/functions-triggers-bindings-overview
    """

    @staticmethod
    def is_supported_trigger_type(trigger_instance: 'Trigger',
                                  trigger_type: Type['Trigger']):
        return isinstance(trigger_instance,
                          trigger_type) or trigger_instance.type == \
            trigger_type.get_binding_name()

    def __init__(self, name: str, data_type: Optional[DataType] = None,
                 type: Optional[str] = None) -> None:
        super().__init__(direction=BindingDirection.IN,
                         name=name, data_type=data_type, type=type)


class InputBinding(Binding, ABC, metaclass=ABCBuildDictMeta):
    """Class representation of Azure Function Input Binding. \n
    Ref: https://aka.ms/functions-triggers-bindings-overview
    """

    def __init__(self, name: str, data_type: Optional[DataType] = None,
                 type: Optional[str] = None) -> None:
        super().__init__(direction=BindingDirection.IN,
                         name=name, data_type=data_type, type=type)


class OutputBinding(Binding, ABC, metaclass=ABCBuildDictMeta):
    """Class representation of Azure Function Output Binding. \n
    Ref: https://aka.ms/functions-triggers-bindings-overview
    """

    def __init__(self, name: str, data_type: Optional[DataType] = None,
                 type: Optional[str] = None) -> None:
        super().__init__(direction=BindingDirection.OUT,
                         name=name, data_type=data_type, type=type)


class Setting(ABC, metaclass=ABCBuildDictMeta):
    """ Abstract class for all settings of a function app.
        This class represents all the decorators that cannot be
        classified as bindings or triggers. e.g function_name, retry etc.
    """

    EXCLUDED_INIT_PARAMS = {'self', 'kwargs', 'setting_name'}

    def __init__(self, setting_name: str) -> None:
        self.setting_name = setting_name
        self._dict: Dict = {
            "setting_name": self.setting_name
        }

    def get_setting_name(self) -> str:
        return self.setting_name

    def get_dict_repr(self) -> Dict:
        """Build a dictionary of a particular binding. The keys are camel
        cased binding field names defined in `init_params` list and
        :class:`Binding` class. \n
        This method is invoked in function :meth:`get_raw_bindings` of class
        :class:`Function` to generate json dict for each binding.

        :return: Dictionary representation of the binding.
        """
        params = list(dict.fromkeys(getattr(self, 'init_params', [])))
        for p in params:
            if p not in Setting.EXCLUDED_INIT_PARAMS:
                self._dict[p] = getattr(self, p, None)

        return self._dict

    def get_settings_value(self, settings_attribute_key: str) -> Optional[str]:
        """
        Get the value of a particular setting attribute.
        """
        return self.get_dict_repr().get(settings_attribute_key)
